#!/usr/bin/env python3
#_*_ coding:utf-8 _*_

import statistics, sys
import math
import ahocorasick
from .diPwm import diPWM, create_diPwm
from .SemiNaive import search_semi_naive_LAM_window, search_semi_naive_LAM


def compute_positions_kept(diP, k):
	""" Compute the list of positions of the diPWM object where the standard deviation is more than `k`.

	Args:
		diP (diPWM): object diPWM

		k (float): threshold given to select the position of the diPWM object

	Return:
		list of int: a list of positions of the diPWM object
	"""

	list_positions_kept = []
	for i in range(diP.length):
		dico = diP.value[i]
		standard_dev = statistics.stdev(list(dico.values()))
		if standard_dev > k :
			list_positions_kept.append(i)
	print(f'Liste des positions gardées : {list_positions_kept}', file=sys.stderr)
	return list_positions_kept


def make_blocks(list_positions_kept):
	""" Translate a list of position in a list of blocks.

	Args:
		list_positions_kept (list of int): list of positions

	Return:
		list of int: a list of blocks
	"""

	list_blocks = []
	for i in range(len(list_positions_kept)):
		if i == 0 or list_positions_kept[i] != list_positions_kept[i - 1] + 1:
			list_blocks.append([])
			list_blocks[-1].append(list_positions_kept[i])
		else:
			list_blocks[-1].append(list_positions_kept[i])
	print(f'Liste des blocs : {list_blocks}', file=sys.stderr)
	return list_blocks


def get_longest_block(list_blocks):
	""" Give the longest block of a list of blocks.

	Args:
		list_blocks (list of int): list of blocks

	Return:
		list of int: list of successive positions (a block)
	"""

	list_length = []
	for block in list_blocks:
		list_length.append(len(block))
	index_longest_block = (-max((x,-i) for i,x in enumerate(list_length))[1])
	print(f'Plus long bloc : {list_blocks[index_longest_block]}, taille bloc : {len(block)}', file=sys.stderr)
	return list_blocks[index_longest_block]


def enumerate_block(diP, block, threshold):
	""" This function enumerates words of length of `block + 1` which scores are in the total diPWM >= to `threshold`.
	To estimate score, this function uses LAM and LBM to extend the score on the right and on the left.

	Args:
		diP (diPWM): object diPWM

		block (list of int): list of successive positions (a block)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: word, score
	"""

	LBM = diP.make_look_back_matrix()
	stack = []

	# for root, child and parent : [0]:level, [1]:score, [2]:prefix
	# one root per symbol of the alphabet
	for nt in diP.alphabet:
		root = (block[0], LBM[block[0]][nt],nt)
		stack.append(root)

	while stack:
		parent = stack.pop()

		for d in diP.alphabet:
			child = (
				parent[0] + 1,
				parent[1] + diP.value[parent[0]][parent[2][-1] + d],
				parent[2] + d
			)
			if child[1] + diP.LAM[child[0]][d] >= threshold :
				if parent[0] == block[-1] :
					yield child[2], child[1] + diP.LAM[child[0]][d]		# generate word, score
				else :
					stack.append(child)


def search_block(diP, text, threshold):
	""" Search of a set of subwords through a text using Aho-Corasick algorithm for a given threshold and check the windows by using LAM.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: starting_position, word, score
	"""

	list_positions_kept = compute_positions_kept(diP,1)
	list_blocks = make_blocks(list_positions_kept)
	block = get_longest_block(list_blocks)
	autom = ahocorasick.Automaton()
	for word, score in enumerate_block(diP, block, threshold):
		autom.add_word(word, (word,score))
	autom.make_automaton()

	for (position, (word, score)) in autom.iter(text):
		start_position = position - (block[-1] + 1)
		if position < len(text) - diP.length + block[-1]:
			score = search_semi_naive_LAM_window(diP, start_position ,text, threshold)
			if score:
				yield start_position,text[start_position:start_position+diP.length], score


def search_block_ratio(diP, text, ratio):
	""" Search of a set of subwords through a text using Aho-Corasick algorithm for a given threshold and check the windows by using LAM. From the ratio is calculated the threshold.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: starting_position, word, score
	"""

	# calculate threshold from ratio
	threshold = diP.set_threshold_from_ratio(ratio)
	for position, word, score in search_block(diP, text, threshold):
		yield position, word, score


def get_block_optimized(diP):
	""" Compute a block of the diPWM object by a heuristic.

	Args:
		diP (diPWM): object diPWM

	Return:
		list of int: list of successive positions (a block)
	"""

	li_sd = [statistics.stdev(list(dico.values())) for dico in diP.value]
	li_interval_possible = [(list(range(i,j+1)),statistics.mean(li_sd[i:j+1])) for i in range(len(diP)) for j in range(i,len(diP)) if j - i + 1 >= min(len(diP)-3,10)]
	li_interval_possible.sort(key=lambda x:x[1],reverse=True)
	return li_interval_possible[0][0]


def search_block_optimized(diP, text, threshold):
	""" Search of a set of subwords through a text using Aho-Corasick algorithm for a given threshold and check the windows by using LAM. The subwords are compute with a heuristic.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: starting_position, word, score
	"""
	block = get_block_optimized(diP)

	autom = ahocorasick.Automaton()
	for word, score in enumerate_block(diP, block, threshold):
		autom.add_word(word, (word,score))
	autom.make_automaton()

	for (position, (word, score)) in autom.iter(text):
		start_position = position - (block[-1] + 1)
		if position < len(text) - diP.length + block[-1]:
			score = search_semi_naive_LAM_window(diP, start_position ,text, threshold)
			if score:
				yield start_position,text[start_position:start_position+diP.length], score


def search_block_optimized_ratio(diP, text, ratio):
	""" Search of a set of subwords through a text using Aho-Corasick algorithm for a given threshold and check the windows by using LAM. The subwords are compute with a heuristic. From the ratio is calculated the threshold.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: starting_position, word, score
	"""
	# calculate threshold from ratio
	threshold = diP.set_threshold_from_ratio(ratio)
	for position, word, score in search_block_optimized(diP, text, threshold):
		yield position, word, score


#



#!/usr/bin/env python3
#_*_ coding:utf-8 _*_

import math

from .diPwm import diPWM
from .Enumerate import enumerate_words_LAM

def get_first_threshold(diP, number_words_wanted):
	""" Compute a first threshold by using uniforme probability.

	Args:
		diP (diPWM): object diPWM

		number_words_wanted (int): number of words that we want to compute

	Return:
		list of int: list of successive positions (a block)
	"""

	# possibles = diP.alphabet_size^(diP.length + 1)
	possibles = math.pow(diP.alphabet_size, diP.length + 1)
	ratio = number_words_wanted/possibles
	threshold = diP.max - ((diP.max - diP.min)*ratio)
	return threshold

def count_words(diP, threshold, range_count):
	""" Compute the number of words enumerated where it is less or equal than `range_count`. If this number is more than `range_count`, the function returns `range_count + 1`.

	Args:
		diP (diPWM): object diPWM

		threshold (float): threshold given to select the windows

		range_count (int): limit of the number of words that we want to compute

	Return:
		int: number of words enumerate
	"""

	number = 0
	for word, score in enumerate_words_LAM(diP,threshold):
		number += 1
		if number == range_count + 1:
			return number
	return number


def find_threshold_dichotomy(diP, number_words_wanted, range_count, bound_min = None, bound_max = None):
	""" Compute a threshold where the number of words enumerated is between `number_words_wanted` and `range_count`.

	Args:
		diP (diPWM): object diPWM

		number_words_wanted (int): number of words that we want to compute

		range_count (int): limit of the number of words that we want to compute

		bound_min (str, optional): bound minimum for the dichotomy. Defaults to the minimum of diP.

		bound_max (str, optional): bound maximum for the dichotomy. Defaults to the maximum of diP.

	Return:
		int: a threshold where the number of words enumerated is between `number_words_wanted` and `range_count`
	"""

	if bound_min == None:
		bound_min = diP.min
	if bound_max == None:
		bound_max = diP.max
	threshold = (bound_min + bound_max)/2
	number_words = count_words(diP, threshold, range_count)
	while not number_words_wanted <= number_words <= range_count :
		if number_words < number_words_wanted:
			bound_max = threshold
		else:
			bound_min = threshold
		threshold = (bound_min + bound_max)/2
		number_words = count_words(diP, threshold, range_count)
	return threshold

def find_threshold_dichotomy_optimized(diP, number_words_wanted, range_count):
	""" Compute a threshold where the number of words enumerated is between `number_words_wanted` and `range_count`. In this optimized version, we use the function `get_first_threshold` to compute the beginning of the dichotomy.

	Args:
		diP (diPWM): object diPWM

		number_words_wanted (int): number of words that we want to compute

		range_count (int): limit of the number of words that we want to compute

	Return:
		int: a threshold where the number of words enumerated is between `number_words_wanted` and `range_count`
	"""

	threshold = get_first_threshold(diP, number_words_wanted)
	number_words = count_words(diP, threshold, range_count)
	bound_max = diP.max
	bound_min = threshold
	while not number_words > range_count:
		threshold -= (bound_max - bound_min)
		bound_min = threshold
		number_words = count_words(diP, threshold, range_count)
	threshold = find_threshold_dichotomy(diP, number_words_wanted, range_count, bound_min, bound_max)
	return threshold

def find_best_words(diP, number_words_wanted):
	""" Compute the `number_words_wanted` enumerated words by decreasing score.

	Args:
		diP (diPWM): object diPWM

		number_words_wanted (int): number of words that we want to compute

	Return:
		list of tuple: list of the first `number_words_wanted` (enumerated word, score) by decreasing score
	"""

	threshold = find_threshold_dichotomy_optimized(diP, number_words_wanted, number_words_wanted*2)
	li = list(enumerate_words_LAM(diP,threshold))
	li.sort(key=lambda x: x[1], reverse=True)
	return li[:number_words_wanted]



#!/usr/bin/env python3
#_*_ coding:utf-8 _*_
from src.diPwm import diPWM, create_diPwm
from src.BestWords import get_first_threshold, count_words, find_threshold_dichotomy, find_threshold_dichotomy_optimized, find_best_words

def test_first_threshold():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diPwm_matrix_new = create_diPwm(pathDiPwm)
    c = get_first_threshold(diPwm_matrix_new, 1000)

    assert round(c,6) == 21.663316

def test_count_words():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diPwm_matrix_new = create_diPwm(pathDiPwm)
    c = get_first_threshold(diPwm_matrix_new, 1000)
    sol = count_words(diPwm_matrix_new, c, 2000)

    assert sol == 1

def test_find_threshold_dichotomy():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diPwm_matrix_new = create_diPwm(pathDiPwm)
    threshold = find_threshold_dichotomy(diPwm_matrix_new, 1000, 2000)

    assert round(threshold,6) == 14.606681

def test_find_threshold_dichotomy_optimized():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diPwm_matrix_new = create_diPwm(pathDiPwm)
    threshold = find_threshold_dichotomy_optimized(diPwm_matrix_new, 1000,2000)

    assert round(threshold,6) == 14.772161

def test_find_best_words():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diPwm_matrix_new = create_diPwm(pathDiPwm)
    sol = find_best_words(diPwm_matrix_new,10)

    assert [(x,round(y,6)) for x,y in sol] == [('CGGTCACGTGAC', 21.667164), ('GGGTCACGTGAC', 21.577231), ('TGGTCACGTGAC', 21.507651), ('CGGTCACGTGAG', 21.415541), ('GGGTCACGTGAG', 21.325608), ('TGGTCACGTGAG', 21.256028), ('CGGTCACGTGAT', 21.239144), ('AGGTCACGTGAC', 21.226616), ('GGGTCACGTGAT', 21.149211), ('TGGTCACGTGAT', 21.079631)]
